/*******************************************************************************
 * 
 * Arquivo que manipula os temporizadores que sinalizam deadlines expirados para 
 * o controle de quanto de processamento é disponibilizado para cada processo.
 *
 ******************************************************************************/

#include "alarms.h"
#include "macros.h"

int global_sockfd;
extern int numberOfProcessors;

/** Tratador de notificações geradas pelo monitoramento dos processos de tempo 
real */
void alarmHandler(int sig)
{
    int res;
    pthread_attr_t attr;

    pthread_t alarmThread;

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    pthread_attr_setinheritsched(&attr, PTHREAD_INHERIT_SCHED);

    res = pthread_create(&alarmThread, &attr, treatAlarms, NULL);
    if (res != 0) {
        printf("Não foi possível criar thread para o acompanhamento de processos\n");
        exit(1);
    }
}

/** Tratador de notificações geradas pelo monitoramento dos processos de tempo 
real */
void *treatAlarms()
{
    long int executionTime;
    long int finishSliceTime;
    struct timeval timeNow;
    unsigned long now;
    unsigned long remainingPeriod;
    float sliceDivisor;

    struct processDataStruct *node;
    struct alarmDataStruct alarm;
    int hasEqualAlarms;

#ifdef DEBUG
    printf("---------------------------------------------------------------\n");
    printf("AlarmHandler");
#endif

    while (1) {
        hasEqualAlarms = removeFromAlarmListHead(&alarm);
        node = alarm.process;

        /* XXX verificar se o tempo de processamento do processo está ok, senão bloqueá-lo até o próximo período! */
        gettimeofday(&timeNow, NULL);
        /* Tempo em milisegundos desde a meia-noite de 01 de janeiro de 1970 */
        now = (timeNow.tv_sec * 1000) + (timeNow.tv_usec / 1000);

        executionTime = getElapsedTime(node);
#ifdef DEBUG
        printf(" em %lu\n", now);
        printf("Processo %d\n", node->pid);
        printf("Tempo total de execucao: %lu\n", executionTime);
#endif

        if (executionTime == -1) {
            /* Processo terminou */

            printf("Processo finalizado!\n");
            /* Retira processo da lista de processos */
            removeClientFromList(node->pid);
        } else {
            /* Process ainda não terminou */
#ifdef DEBUG
            printf("Processo ainda não terminou.\n");
#endif
            sliceDivisor = MIN(node->num_child, numberOfProcessors);

            if (node->running == 0) {
                /* Processo ainda não terminou
                 * Processo não está rodando */

                wakeUpProcessTree(node);
                node->running = 1;
                node->next_period += PERIOD;
#ifdef DEBUG
                printf("Processo foi despertado para executar sua fatia.\n");
                printf("Fim do período será em: %lu.\n", node->next_period);
#endif
                insertAtAlarmList(node, iround(node->slice / sliceDivisor));
            } else {
                /* Processo ainda não terminou
                 * Processo em execução */

                unsigned long elapsedSlice = executionTime - node->executed_time;
#ifdef DEBUG
                printf("Processo em execução.\n");
                printf("Tempo de exec decorrido no periodo: %lu.\n", elapsedSlice);
#endif
                if (elapsedSlice >= node->slice) {
                    /* Processo em execução
                     * Processo executou sua fatia */
#ifdef DEBUG
                    printf("Processo já executou sua fatia.\n");
#endif
                    if (now < node->next_period) {
                        /* Processo em execução
                         * Processo executou sua fatia
                         * Período não acabou */
#ifdef DEBUG
                        printf("Parando processo até o fim do período.\n");
#endif
                        /* Parando processo */
                        stopProcessTree(node);
                        node->running = 0;
                        node->executed_time = executionTime;

                        insertAtAlarmList(node, (int) (node->next_period - now));
                    } else {    /* now >= node->next_period */
                        /* Processo em execução
                         * Processo executou sua fatia   
                         * Período terminou */

                        node->next_period += PERIOD;
                        node->executed_time = executionTime;
#ifdef DEBUG
                        printf("Período e fatia esgotados.\n");
                        printf("Fim do próximo período: %lu\n", node->next_period);
#endif
                        insertAtAlarmList(node, iround(node->slice / sliceDivisor));
                    }
                } else {        /* elapsedSlice < slice */
                    /* Processo em execução
                     * Processo ainda não executou sua fatia */
#ifdef DEBUG
                    printf("Processo não executou slice time ainda.\n");
#endif
                    if (node->next_period < now) {
                        /* (unsigned long) */
                        remainingPeriod = 0;
                    } else {
                        remainingPeriod = node->next_period - now;
                    }
                    if (remainingPeriod / sliceDivisor >= JIFFY) {
                        /* Processo em execução
                         * Processo ainda não executou sua fatia
                         * Período ainda não acabou */

                        /* Agendando próximo alarme */
                        finishSliceTime = now + iround((node->slice - elapsedSlice) / sliceDivisor);
                        if (finishSliceTime < node->next_period) {
#ifdef DEBUG
                            printf("Agendando para o fim da fatia.\n");
#endif
                            insertAtAlarmList(node,
                                              iround((node->slice -
                                                      elapsedSlice) / (float) node->num_child));
                        } else {        /* finishSliceTime >= node->next_period */
#ifdef DEBUG
                            printf("Executará até o fim do período apenas parte da fatia.\n");
#endif
                            insertAtAlarmList(node, (int) (node->next_period - now));
                        }
                    } else {    /* now >= node->next_period */
                        /* Processo em execução
                         * Processo ainda não executou sua fatia
                         * Período terminou */

                        node->next_period += PERIOD;
                        node->executed_time = executionTime;
#ifdef DEBUG
                        printf("Período esgotado sem terminar a fatia!!!\n");
                        printf("Fim do próximo período: %lu\n", node->next_period);
#endif
                        insertAtAlarmList(node, iround(node->slice / (float) node->num_child));
                    }
                }
            }
        }
        if (hasEqualAlarms == 0)
            break;
    }
    return NULL;
}

/*
 * Thread que trata as conexões que chegam ao servidor.
 * 
 */
void *treatIOs()
{
    int sockfd;
    int res;
    pthread_attr_t attr;

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    for (;;) {
        pthread_t msgThread;

        printf("\n***\nEsperando conexões\n***\n");
        fflush(stdout);
        sockfd = acceptConnection(global_sockfd);
        if (sockfd < 0) {
            printf("accept será chamado novamente.\n");
            continue;
        }
        printf("\n***\nConexão aceita\n***\n");
        fflush(stdout);

        res = pthread_create(&msgThread, &attr, treatMsgs, (void *) (intptr_t) sockfd);
        if (res != 0)
            printf("Não foi possível criar thread para tratamento do processo cliente");
    }

    return NULL;
}

/** Thread que manipula msgs enviadas pelos clientes e as direciona para 
* tratamento de reserva/adaptação
*/
void *treatMsgs(void *sockfd)
{
    struct serverAdaptMsgDataStruct serverAdaptMsgData;
    struct reserveMsgDataStruct reserveMsgData;
    char msg[256];
    pid_t pid;
    char *status;

    msg[0] = '\0';
    recvClientOption((int) (intptr_t) sockfd, msg);

    if (strcmp(msg, RESERVE) == 0) {
        sndMsgClient((int) (intptr_t) sockfd, "", 1);
        recvReserveMsg((int) (intptr_t) sockfd, &reserveMsgData);
        pid = createProcess(reserveMsgData.command, reserveMsgData.params);
        reserveMsgData.pid = pid;
        sprintf(msg, "%d", pid);
        sndMsgClient((int) (intptr_t) sockfd, msg, strlen(msg) + 1);
        treatReserveMessage(&reserveMsgData);
    } else if (strcmp(msg, SERVER_ADAPT) == 0) {
        sndMsgClient((int) (intptr_t) sockfd, "Adaptação do servidor solicitada", 35);
        recvServerAdaptMsg((int) (intptr_t) sockfd, &serverAdaptMsgData);
        sndMsgClient((int) (intptr_t) sockfd, "Adaptação do servidor aceita.", 32);
        treatServerAdaptMessage(&serverAdaptMsgData);
    } else if (strcmp(msg, SERVER_STATUS) == 0) {
        status = server_status();
        sndMsgClient((int) (intptr_t) sockfd, status, strlen(status) + 1);
        free(status);
    } else if (strcmp(msg, KILL_PID) == 0) {
        /* Enviando dummy msg porque sendClientOption espera */
        sendACK((int) (intptr_t) sockfd);
        pid = recvKillMsg((int) (intptr_t) sockfd);
        killProcessTree(pid);
    }


    /* Fecha a conexão com o cliente */
    closeSocket((int) (intptr_t) sockfd);

    return NULL;
}

/** Tratador de interrupção do teclado:CTRL^C - finaliza servidor. */
void intHandler(int sig)
{
    killProcesses();
    exit(0);
}

/** Tratador do término de um processo: só espera processo finalizar para que o 
* mesmo não se torne um zumbi.
*/
void childHandler(int sig)
{
    int status;
    int res;

    res = waitpid(-1, &status, WNOHANG);
    if (res < 0) {
        return;
    }
}

/** Cria os alarmes temporizadores, de interrupção de teclado e de finalização 
* de processos
*/
void createAlarms(pid_t pid, int sockfd)
{
    struct sigaction saalarm;
    struct sigaction saint;
    struct sigaction sachild;
    global_sockfd = sockfd;

    sigemptyset(&saalarm.sa_mask);
    saalarm.sa_handler = alarmHandler;
    saalarm.sa_flags = 0;
    sigaction(SIGUSR1, &saalarm, NULL);
    if (createTimer() == -1)
        error("Erro ao criar temporizador");

    sigemptyset(&saint.sa_mask);
    saint.sa_handler = intHandler;
    saint.sa_flags = 0;
    sigaction(SIGINT, &saint, NULL);

    sigemptyset(&sachild.sa_mask);
    sachild.sa_handler = childHandler;
    sachild.sa_flags = SA_NOCLDSTOP;
    sigaction(SIGCHLD, &sachild, NULL);

    if (createListsSemaphores() == -1)
        error("Erro no sem_init das listas");
    if (createSocketSemaphore() == -1)
        error("Erro no sem_init dos sockets");
}
