/*******************************************************************************
 * 
 * Arquivo contendo métodos para a manipulação de listas ligadas. Existem duas 
 * listas, uma contendo processos sendo executados e outra contendo o próximo 
 * deadline de execução para cada processo.
 *
 ******************************************************************************/

#include"lists.h"

sem_t alarmListSem;
sem_t processListSem;

extern struct processDataStruct *processList;
extern sliceListT sliceList;

/** Adiciona um processo à lista local de processos de tempo real */
struct processDataStruct *addClientToList(pid_t pid)
{
    extern struct processDataStruct *processList;
    struct processDataStruct *new;
    struct timeval timeNow;
    unsigned long now;
    struct pidDataStruct *father;

    new = (struct processDataStruct *)
        malloc(sizeof(struct processDataStruct));
    new->pid = pid;
    father = (struct pidDataStruct *) malloc(sizeof(struct pidDataStruct));
    father->pid = pid;
    father->next = NULL;
    new->childrenList = father;
    new->num_child = 1;
    sem_init(&new->childListSem, 0, 1);

    gettimeofday(&timeNow, NULL);
    /* Tempo em milisegundos desde a meia-noite de 01 de janeiro de 1970 */
    now = (timeNow.tv_sec * 1000) + (timeNow.tv_usec / 1000);
    new->next_period = now + PERIOD;
    new->executed_time = 0;
    new->running = 1;

    sem_wait(&processListSem);
    new->next = processList;
    processList = new;
    addAppToSliceList(sliceList, new->pid, new->num_child);
    updateSlices(processList, sliceList);
    sem_post(&processListSem);

#ifdef DEBUG
    printf("Adicionando processo em %lu\n", now);
    printf("Fim do período será em: %lu\n", new->next_period);
#endif

    return new;
}

/** Remove um processo da lista local de processos de tempo real */
void removeClientFromList(pid_t pid)
{
    extern struct processDataStruct *processList;
    struct processDataStruct *prev_node;
    struct processDataStruct *node;

    sem_wait(&processListSem);
    if (processList != NULL) {
        node = processList;
        if (node->pid == pid) {
            processList = node->next;
            free(node);
        } else {
            prev_node = node;
            node = node->next;
            while (node != NULL) {
                if (node->pid == pid) {
                    prev_node->next = node->next;
                    free(node);
                    break;
                }
                prev_node = node;
                node = node->next;
            }
        }
    }
    /* Aumenta fatia dos demais */
    sem_post(&processListSem);
    removeAppFromSliceList(sliceList, pid);
    updateSlices(processList, sliceList);
    removeFIFOs(pid);
}

/** Mata todos os processos da lista de processos */
void killProcesses()
{
    extern struct processDataStruct *processList;
    struct processDataStruct *node;

    sem_wait(&processListSem);
    node = processList;
    while (node != NULL) {
        printf("\nVou matar processo %d", node->pid);
        kill(node->pid, SIGKILL);
        removeFIFOs(node->pid);
        processList = node->next;
        free(node);
        node = processList;
    }
    sem_post(&processListSem);
}

/* Mata a árvore toda de um processo */
void killProcessTree(pid_t pid)
{
    struct processDataStruct *tree;
    struct pidDataStruct *node;

    tree = getProcessFromPid(pid);
    if (tree == NULL)
        return;
    node = tree->childrenList;

    sem_wait(&(tree->childListSem));
    killProcessTreeRec(node);
    sem_post(&(tree->childListSem));
}

/* Mata processos filhos recursivamente */
void killProcessTreeRec(struct pidDataStruct *node)
{
    if (node->next != NULL)
        killProcessTreeRec(node->next);
    kill(node->pid, SIGKILL);
}

void insertAtAlarmList(struct processDataStruct *process, unsigned long time)
{
    /* Inserção linear */
    extern struct alarmDataStruct *alarmList;
    struct alarmDataStruct *new;
    struct alarmDataStruct *node;
    struct alarmDataStruct *prev_node;
    struct timeval timeNow;
    unsigned long alarmTime;
    unsigned long now;

    gettimeofday(&timeNow, NULL);
    now = (timeNow.tv_sec * 1000) + (timeNow.tv_usec / 1000);
    alarmTime = now + time;

    sem_wait(&alarmListSem);
#ifdef DEBUG
    printf("Inserindo %d %lu em %lu. Próximo timer eh em %lu\n",
           process->pid, time, now, alarmTime);
#endif

    new = (struct alarmDataStruct *) malloc(sizeof(struct alarmDataStruct));

    new->alarmTime = alarmTime;
    new->process = process;
    new->next = NULL;

    node = alarmList;
    prev_node = NULL;

    if (alarmList == NULL)
        alarmList = new;
    else {
        while (node != NULL) {
            if (node->alarmTime > alarmTime) {
                new->next = node;
                if (alarmList == node) {
                    alarmList = new;
                } else {
                    prev_node->next = new;
                }
                break;
            }
            prev_node = node;
            node = node->next;
        }
        if (node == NULL)
            prev_node->next = new;
    }
    /* O novo alarme é o que será disparado mais cedo.  */
    if (new == alarmList) {
#ifdef DEBUG
        printf("Pelo insert em %lu, próximo alarm eh: %d %lu\n", now,
               new->process->pid, new->alarmTime);
#endif
        if (setTimer(0, 0, 0, time * 1000) == -1) {
            printf("Erro no setTimer...\n");
        }
    }
    sem_post(&alarmListSem);
}

void removeFromAlarmList(struct processDataStruct *process)
{
    extern struct alarmDataStruct *alarmList;
    struct alarmDataStruct *alarm;
    struct alarmDataStruct *prev_alarm;
    int remove = 0;

    sem_wait(&alarmListSem);
    if (alarmList != NULL) {
        alarm = alarmList;
        if (alarm->process->pid == process->pid) {
            if (setTimer(0, 0, 0, 0) == -1)
                printf("\nErro no setTimer...");
            remove = 1;
        } else {
            prev_alarm = alarm;
            alarm = alarm->next;
            while (alarm != NULL) {
                if (alarm->process->pid == process->pid) {
                    prev_alarm->next = alarm->next;
                    free(alarm);
                    break;
                }
                prev_alarm = alarm;
                alarm = alarm->next;
            }
        }
    }
    sem_post(&alarmListSem);

    if (remove == 1)
        removeFromAlarmListHead(alarm);
}

int removeFromAlarmListHead(struct alarmDataStruct *alarm)
{
    extern struct alarmDataStruct *alarmList;
    struct alarmDataStruct *data;
    struct timeval timeNow;
    unsigned long now;
    unsigned long time;
    int result;

    sem_wait(&alarmListSem);
    data = alarmList;
    alarm->process = data->process;
    alarmList = alarmList->next;

    if (alarmList != NULL) {
        gettimeofday(&timeNow, NULL);
        now = (timeNow.tv_sec * 1000) + (timeNow.tv_usec / 1000);

#ifdef DEBUG
        printf("\nPelo remove em %lu, próximo alarm eh: %d %lu", now,
               alarmList->process->pid, alarmList->alarmTime);
#endif
        time = alarmList->alarmTime - now;
        if (alarmList->alarmTime > now) {
            if (setTimer(0, 0, 0, time * 1000) == -1)
                printf("\nErro no setTimer...");
            result = 0;
        } else {
            result = 1;
        }

    } else {
        result = 0;
    }
    free(data);
    sem_post(&alarmListSem);

    return result;
}

int checkProcessExistance(pid_t pid)
{
    struct processDataStruct *proc = NULL;

    proc = getProcessFromPid(pid);
    if (proc == NULL)
        return -1;
    else
        return 1;
}

struct processDataStruct *getProcessFromPid(pid_t pid)
{
    extern struct processDataStruct *processList;
    struct processDataStruct *node = NULL;

    sem_wait(&processListSem);
    node = processList;
    while (node != NULL) {
        if (node->pid == pid) {
            break;
        } else {
            node = node->next;
        }
    }
    sem_post(&processListSem);

    return node;
}

int createListsSemaphores()
{
    if (sem_init(&processListSem, 0, 1) == -1)
        return -1;
    if (sem_init(&alarmListSem, 0, 1) == -1)
        return -1;
    return 1;
}
