/*******************************************************************************
 * 
 * Arquivo contendo métodos que implementam a manipulação dos processos ativos
 * para que os mesmos sejam executados nos períodos de tempo pré-determinados.
 *
 ******************************************************************************/

#include"reserve.h"

sem_t alarmListSem;
sem_t processListSem;
extern cpu_set_t cpuSet;
extern struct processDataStruct *processList;
extern int processingLimit;
extern int numberOfProcessors;
struct reservationDataStruct *reservationList = NULL;

/** Método para processamento de mensagens recebidas dos clientes. Após o 
processamento, a conexão por onde os dados chegaram é fechada. */
void treatReserveMessage(struct reserveMsgDataStruct *msgData)
{
    struct processDataStruct *process;

    process = addClientToList(msgData->pid);
    startMonitorClient(process);
}

/** Monitora a execução dos processos. O objetivo é identificar processos 
com tempo de CPU maior que a especificada na admissão do processo */
void startMonitorClient(struct processDataStruct *process)
{
#ifdef DEBUG
    printf("Reiniciando o processo %d...\n", process->pid);
#endif
    wakeUpProcessTree(process);

#ifdef DEBUG
    printf("Dando inicio ao monitoramento do processo %d\n", process->pid);
#endif
    insertAtAlarmList(process, process->slice);
}

void stopMonitorClient(struct processDataStruct *process)
{
#ifdef DEBUG
    printf("Parando o monitoramento do processo %d\n", process->pid);
#endif

    removeFromAlarmList(process);
}

int createTimer()
{
    int err;
    struct sigevent signal_specification;

    signal_specification.sigev_signo = SIGUSR1;

    err = timer_create(CLOCK_REALTIME, &signal_specification, &created_timer);
    if (err != 0)
        return -1;
    return 1;
}

int setTimer(long interval_sec, long interval_usec, long value_sec, long value_usec)
{
    struct itimerspec new_setting;
    int err;

    if (value_usec * 1000 >= 1000000000) {
        value_sec += (value_usec / 1000000);
        value_usec = (value_usec % 1000000);
    }

    new_setting.it_value.tv_sec = value_sec;
    new_setting.it_value.tv_nsec = value_usec * 1000;
    new_setting.it_interval.tv_sec = 0;
    new_setting.it_interval.tv_nsec = 0;

    err = timer_settime(created_timer, 0, &new_setting, NULL);
    if (err != 0) {
        return -1;
    }
    return 1;
}

void wakeUpProcessTree(struct processDataStruct *process)
{
    struct pidDataStruct *childrenList;

    childrenList = process->childrenList;
    while (childrenList != NULL) {
        kill(childrenList->pid, SIGCONT);
        childrenList = childrenList->next;
    }
}

void stopProcessTree(struct processDataStruct *process)
{
    struct pidDataStruct *childrenList;

    childrenList = process->childrenList;
    while (childrenList != NULL) {
        kill(childrenList->pid, SIGSTOP);
        childrenList = childrenList->next;
    }
}

/*
 * Funções para alterar a afinidade dos processos
 */

/** Amarra a execução de um processo a um conjunto de processadores específico. */
int setProcessAffinity(pid_t pid)
{
    extern cpu_set_t cpuSet;

    if (sched_setaffinity(pid, sizeof(cpuSet), &cpuSet) < 0)
        return -1;
    return 1;
}

/** Atualiza em quais processadores os processos rodam */
void updateProcessesAffinity()
{
    struct processDataStruct *process;
    struct pidDataStruct *child;

    for (process = processList; process != NULL; process = process->next) {
        for (child = process->childrenList; child != NULL; child = child->next) {
            if (setProcessAffinity(child->pid) == -1) {
                printf("\nNão foi possível configurar afinidade de processador. Continuando...");
            }
        }
    }
}

void restartProcessMonitor(struct processDataStruct *process)
{
    struct timeval timeNow;
    unsigned long now;

    process->executed_time = 0;
    process->running = 1;

    gettimeofday(&timeNow, NULL);
    /* Tempo em milisegundos desde a meia-noite de 01 de janeiro de 1970 */
    now = (timeNow.tv_sec * 1000) + (timeNow.tv_usec / 1000);
    process->next_period = now + PERIOD;

    stopMonitorClient(process);
    startMonitorClient(process);
}

void updateSlices(struct processDataStruct *processList, sliceListT sliceList)
{
    struct processDataStruct *p;
    short int factor;
    unsigned long min, max;

    calcSlices(sliceList);
    for (p = processList; p; p = p->next) {
        factor = MIN(p->num_child, numberOfProcessors);
        min = JIFFY * factor;
        max = RESERVATION_LIMIT * factor * PERIOD;

        p->slice = getSliceFromSliceList(sliceList, p->pid);
        if (p->slice < min)
            p->slice = min;
        else if (p->slice > max)
            p->slice = max;
    }
}
