/*******************************************************************************
 * 
 * Arquivo contendo a main do servidor. Esse arquivo invoca métodos para o 
 * gerencimento dos processos cujos pids foram recebidos como parâmetros.
 *
 ******************************************************************************/

#include"server.h"

struct processDataStruct *processList = NULL;
struct alarmDataStruct *alarmList = NULL;
int numberOfProcessors = 0;
cpu_set_t cpuSet;
int processingLimit;            /* 0 a 100 por simplicação para o usuário */
int mask;

extern sem_t processListSem;
sliceListT sliceList;

/** Método para tratar condições de erro, imprimindo mensagem e terminando o 
programa */
void error(char *msg)
{
    perror(msg);
    exit(1);
}

int calcMaxMask()
{
    int i;
    int maxMask = 0;

    for (i = 0; i < NUMBER_OF_PROCESSORS; i++) {
        maxMask += pow(2, i);
    }

    return maxMask;
}

int makeCpuSet()
{
    int maxMask = 0;
    int nProcessors = 0;
    int i;

    maxMask = calcMaxMask();
    if ((mask > maxMask) || (mask <= 0))
        mask = maxMask;

    CPU_ZERO(&cpuSet);
    for (i = 0; i < sizeof(mask) * 8; i++) {
        if ((mask & (1 << i)) != 0) {
            nProcessors++;
            CPU_SET(i, &cpuSet);
        }
    }

    return nProcessors;
}

void setProcessingLimit(int newProcessingLimit)
{
    int max = iround(RESERVATION_LIMIT * 100);
    int min = ceil(JIFFY / (float) PERIOD);

    if (newProcessingLimit > max)
        newProcessingLimit = max;
    else if (newProcessingLimit < min)
        newProcessingLimit = min;

    processingLimit = newProcessingLimit;
}

void treatServerAdaptMessage(struct serverAdaptMsgDataStruct *msgData)
{
    int oldMask;
    int oldNumberOfProcessors;
    int oldProcessingLimit;

    oldMask = mask;
    oldNumberOfProcessors = numberOfProcessors;
    oldProcessingLimit = processingLimit;

    /* Atualiza numberOfProcessors e mask */
    mask = msgData->mask;
    numberOfProcessors = makeCpuSet();
    setProcessingLimit(msgData->processingLimit);

    if (oldMask != mask) {
        /* Atualiza afinidade do servidor e dos processos cliente */
        setProcessAffinity(getpid());
        updateProcessesAffinity();
    }
    if (oldNumberOfProcessors != numberOfProcessors || oldProcessingLimit != processingLimit) {
        sem_wait(&processListSem);
        setSliceListParams(sliceList, numberOfProcessors, processingLimit);
        updateSlices(processList, sliceList);
        sem_post(&processListSem);
    }
}

int main(int argc, char **argv)
{
    int portNumber;
    pid_t pid;
    int sockfd;
    pthread_t monProcTreeThread;
    pthread_t ioThread;
    pthread_attr_t attr;
    int res;
    char msg[64];
    int i;
    const int alloc_size = 34 * 1024 * 1024;
    char *memory;
    int page_size;

    if (argc < 3) {
        sprintf(msg, "ERRO, usage: %s <port> <CPU usage [1-100]>", argv[0]);
        error(msg);
    }

    portNumber = atoi(argv[1]);
    processingLimit = atoi(argv[2]);
    pid = getpid();
    mask = calcMaxMask();
    numberOfProcessors = makeCpuSet();

    if (setProcessAffinity(pid) == -1) {
        printf("\nNão foi possível configurar afinidade de processador. Continuando...");
    }
    if ((processingLimit < 1)
        || (processingLimit > 100)) {
        error
            ("ERRO, o limite de processamento deve ser no máximo 100 e pelo menos 1! Abortando...");
    }

    changeServerPrio();

    /* Para cálculo das fatias */
    sliceList = createSliceList(numberOfProcessors, processingLimit, PERIOD);

    /* Amarra 32 megas de memória do servidor à RAM da máquina. */
    memory = malloc(alloc_size);
    mlock(memory, alloc_size);
    page_size = getpagesize();
    for (i = 0; i < alloc_size; i += page_size)
        memory[i] = 0;

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    res = pthread_create(&monProcTreeThread, &attr, monitorProcessTree, NULL);
    if (res != 0) {
        printf("\nNão foi possível criar thread de monitoração da árvore de processos");
        exit(1);
    }

    sockfd = createSvrSocket(portNumber);

    createAlarms(pid, sockfd);

    /* Thread para tratar as conexões */
    res = pthread_create(&ioThread, &attr, treatIOs, NULL);
    if (res != 0) {
        printf("Não foi possível criar thread de comunicação com o cliente");
        exit(1);
    }

    printf("Servidor inicializado com %d processadores...\n", numberOfProcessors);

    for (;;) {
        pause();
    }

    freeSliceList(sliceList);
    close(sockfd);

    return 0;
}

/*
    Formato:
    CPU: [mask] [total number of CPUs] [limit percentage]
    Processes:
    -[parent_1 pid]
    --[child_1 pid]
    --[child 2 pid]
    -[parent_2 pid]
    ...
*/
char *server_status()
{
    char *status;
    char *temp;
    extern struct processDataStruct *processList;
    struct processDataStruct *process;
    struct pidDataStruct *child;
    float percent;

    percent = 0.0;

#define MAX_STATUS 1024
#define MAX_TEMP_STATUS 32

    status = malloc(MAX_STATUS * sizeof(char));
    temp = malloc(MAX_TEMP_STATUS * sizeof(char));

    snprintf(status, MAX_STATUS, "CPU: %d %d %d\nProcesses:\n",
             mask, (int) NUMBER_OF_PROCESSORS, processingLimit);

    sem_wait(&processListSem);
    for (process = processList; process != NULL; process = process->next) {
        snprintf(temp, MAX_TEMP_STATUS, "-%d\n", process->pid);
        strcat(status, temp);
        if (process->num_child > 1)
            for (child = process->childrenList->next; child; child = child->next) {
                snprintf(temp, MAX_TEMP_STATUS, "--%d\n", child->pid);
                strcat(status, temp);
            }
    }
    sem_post(&processListSem);

    free(temp);
    return status;
}
